% exampleACALOS_user - stimulus generation function of experiment 'exampleACALOS' -
%
% This function is called by afc_main when starting
% the experiment 'exampleACALOS'. It generates the stimuli which
% are presented during the experiment.
% The stimuli must be elements of the structure 'work' as follows:
%
% work.signal = def.intervallen by 2 times def.intervalnum matrix.
%               The first two columns must contain the test signal
%               (column 1 = left, column 2 = right) ...
%
% work.presig = def.presiglen by 2 matrix.
%               The pre-signal.
%               (column 1 = left, column 2 = right).
%
% work.postsig = def.postsiglen by 2 matrix.
%                The post-signal.
%               ( column 1 = left, column 2 = right).
%
% work.pausesig = def.pausesiglen by 2 matrix.
%                 The pause-signal.
%                 (column 1 = left, column 2 = right).
%
% To design an own experiment, e.g., 'myexperiment',
% make changes in this file and save it as 'myexperiment_user.m'.
% Ensure that the referenced elements of structure 'work' are existing.
%
% See also help exampleACALOS_cfg, exampleACALOS_set, afc_main
% Copyright (c) 1999-2013 Stephan Ewert, Universitaet Oldenburg.
%                         Dirk Oetting, Fraunhofer IDMT
% $Revision: 1.30 beta$  $Date: 26.05.2014 $


function exampleACALOS_user

global def
global work
global setup

% translate current frequency to idx
idxFile = find([250 500 1000 2000 4000 6000 8000] == work.exppar1);
filename = setup.signalnames{idxFile,1};

% get calibrationfilename from def.calib_1kHz and def.signalnames
work.currentCalLevel(1) = setup.calib_1kHz(1) + str2double(setup.signalnames{idxFile,2});
work.currentCalLevel(2) = setup.calib_1kHz(2) + str2double(setup.signalnames{idxFile,3});

%signalfile = fullfile('experiments/wav/', filename);
signalfile = [fileparts(which(mfilename)), filesep, 'ACALOS', filesep, filename];

if ~exist(signalfile,'file')
    error('Soundfile not found!')
else % read file
    if ( work.matlabVersionRelease >= 8.04 )
        [tempsignal, temp_fs] = audioread(signalfile);
    else
        [tempsignal, temp_fs] = wavread(signalfile);
    end
    
    % reduce to single channel
    tempsignal = tempsignal(:,1); 

    % check for sampling rate
    if def.samplerate ~= temp_fs
        error('Sampling rate wrong!')
    end

    % error if file is shorter then the desired interval length
    if def.intervallen > length(tempsignal)
        error('Signal too short!')
    end
end

% level adjustment
tuser = tempsignal/rms(tempsignal);

% for calibration set level to 80 dB SPL(CC)
if setup.calib == 1
    % calculate the dB HL value from 80 dB SPL(CC)
    % now you should calibrate the files that you get 
    % 80 dB SPL(CC) on the levelmeter with the HDA200 headphone
    work.expvaract = 80 - setup.RETSPL(idxFile);
    % activate command window output of current signal presentation
    def.debug = 1;
elseif setup.calib == 2
    for idxFreq = 1:length(def.exppar1)
        freq = def.exppar1(idxFreq);
        filename1 = setup.signalnames{idxFreq,1};
        signalfile = [fileparts(which(mfilename)), filesep, 'ACALOS', filesep, filename1];
        [tuser_tmp, temp_fs] = wavread(signalfile);
        tuser_tmp = tuser_tmp(:,1);
        tuser_tmp = tuser_tmp/rms(tuser_tmp);

        peak_level = 20*log10(max(abs(tuser_tmp)));

        currentCalLevelL = setup.calib_1kHz(1) + str2double(setup.signalnames{idxFreq,2});
        currentCalLevelR = setup.calib_1kHz(2) + str2double(setup.signalnames{idxFreq,3});
        currentCalLevel = min(currentCalLevelL ,currentCalLevelR);

        
        max_presentation_level_dB_HL = currentCalLevel-peak_level;
        max_presentation_level_dB_SPL_CC = currentCalLevel-peak_level+setup.RETSPL(idxFreq);
        max_presentation_level_dB_SPL_FF = currentCalLevel-peak_level+setup.RETSPL(idxFreq)+setup.Gf_Gc(idxFreq);
        fprintf('at %4d Hz max level without clipping: %5.1f dB HL [%5.1f dB SPL (CC), %5.1f dB SPL (FF)]\n',freq,max_presentation_level_dB_HL,max_presentation_level_dB_SPL_CC,max_presentation_level_dB_SPL_FF);
    end

end

% display signal information
if def.debug > 0
    fprintf('%s: %s at %.2f dB HL (%.2f dB SPL(CC))',work.condition,filename,work.expvaract,work.expvaract+setup.RETSPL(idxFile));
end

% reduce to desired lenght
tuser = tuser(1:def.intervallen);

% Hanning windows
tuser = tuser .* setup.window;

% apply calibration from dB HL to dB SPL (CC)
signalL = tuser*10^((work.expvaract-work.currentCalLevel(1))/20);
signalR = tuser*10^((work.expvaract-work.currentCalLevel(2))/20);

% pre-, post- and pausesignals (all zeros)
presig = zeros(def.presiglen,def.outputChannels);
postsig = zeros(def.postsiglen,def.outputChannels);
pausesig = zeros(def.pauselen,def.outputChannels);

% make required fields in work
switch work.condition
    case 'nbl'
        used_ears = [1 0];
    case 'nbr'
        used_ears = [0 1];
    case 'nbb'
        used_ears = [1 1];
    otherwise
        used_ears = [1 1];
        disp('unknown condition, using both ears');
end
work.signal = [ signalL*used_ears(1) ...
    signalR*used_ears(2) ];	% left = right (diotic) first two columns holds the test signal (left right)
work.presig = presig;											% must contain the presignal
work.postsig = postsig;											% must contain the postsignal
work.pausesig = pausesig;										% must contain the pausesignal

% eof